//
// Smile Decorder Next18 AC Rev.A exp.01
// Copyright(C)'2016- Nagoden / Desktop Station / Fujigaya2 / MECY 
//
// Smile Decorder Next18をアクセサリコマンドでサーボを動かす
//  アドレスはCV1とCV9
//　アドレス　= CV9*256 + CV1
//　アドレス　300の場合 CV9 = 1 CV1 =44　
//
//  AUX1　サーボ１　ステータス　ｔ
//  AUX2　サーボ１　ステータス　ｃ
//  AUX3  サーボ１
//  AUX4　サーボ２
//  AUX5　サーボ２　ステータス　ｔ
//  AUX6　サーボ２　ステータス　ｃ
//
//

#include <Arduino.h>
#include <string.h>
#include <avr/pgmspace.h> 
#include <avr/eeprom.h>
#include <SoftwareSerial.h>
#include "NmraDcc_x.h"
#include "VarSpeedServo.h" 

/*************************************************
 * 
 * Declaration of Symbols
 * 
 *************************************************/

//#define DEBUG			//リリースのときはコメントアウトすること
#define MAN_VER_NUMBER  01   /* Release Ver CV07 */
#define MAN_ID_NUMBER 108  /* Manufacture ID */


#define DECODER_ADDRESS 3

//Pin Func  Port
#define PIN_AUX1	A1
#define PIN_F0_F	A2
#define PIN_F0_R	A3
#define PIN_AUX2	A4
#define PIN_AUX5	A5
#define PIN_AUX6  0
#define PIN_AUX4  5
#define PIN_AUX3  6
#define MOTOR_PWM_A 11  
#define MOTOR_PWM_B 3 


//#define MOTOR_LIM_MAX 255

#define CV_AUX3Servo 61           // 0 for OFF, 1 for ON.
#define CV_AUX3Servo_speed 62     // 0-255  1：Slowest　0,255　Fastest
#define CV_AUX3Servo_Angle 63     // 0-180　Depends on servo motor
#define CV_AUX4Servo 64           // 0 for OFF, 1 for ON.
#define CV_AUX4Servo_speed 65     // 0-255  1：Slowest　0,255　Fastest
#define CV_AUX4Servo_Angle 66   // 0-180　Depends on servo motor
#define CV_Servo_Delay 69       // 0-255　CV Vale x 20(ms)  Delay Servocontrol (PIN_F0_R) 

#define CV_SWITCH_0 110           //Staus of Accesory  First Adresse 
#define CV_SWITCH_1 111           //Staus of Accesory  Second Adress 

/*************************************************
 * 
 * Declaration of Classes
 * 
 *************************************************/

NmraDcc	 Dcc;
DCC_MSG	 Packet;

/*************************************************
 * 
 * create servo object to control a servo 
 * 
 *************************************************/

VarSpeedServo myservo1;
VarSpeedServo myservo2;

/*************************************************
 * 
 * Declaration of Variables
 * 
 *************************************************/


//Task Schedule
unsigned long gPreviousL1 = 0;
unsigned long gPreviousL2 = 0;

// CV

uint8_t gCV1_SAddr = 3;
uint8_t gCV9_LAddr = 0;

//Servo
uint8_t gCV61_AUX3Servo = 0;            // 0 for OFF, 1 for ON.
uint8_t gCV62_AUX3Servo_speed = 10;     // 0-255  1：Slowest　0,255　Fastest
uint8_t gCV63_AUX3Servo_Angle = 120;    // 0-180　Depends on servo motor
uint8_t gCV64_AUX4Servo = 0;            // 0 for OFF, 1 for ON.
uint8_t gCV65_AUX4Servo_speed = 10;     // 0-255  1：Slowest　0,255　Fastest
uint8_t gCV66_AUX4Servo_Angle = 120;    // 0-180　Depends on servo motor
uint16_t gCV69_Servo_Delay = 150;           // 0-255  Servo Power Dealy Time(0-5100ms)

//ファンクションの変数
uint8_t gF0_FCONTROL = 0;     //ヘッド・テールのライトライト制御　2018/01/01　nagoden
uint8_t gF0_BCONTROL = 0;     //ヘッド・テールのライトライト制御　2018/01/01　nagoden
uint8_t gFuncBits[29] = {0};
uint8_t gCVFuncTable[2] = {0};


struct CVPair{
  uint16_t	CV;
  uint8_t	Value;
};

CVPair FactoryDefaultCVs [] = {
  {CV_ACCESSORY_DECODER_ADDRESS_LSB, 3},
  {CV_ACCESSORY_DECODER_ADDRESS_MSB, 0},
  {CV_29_CONFIG, 0b10000000},
  {CV_AUX3Servo, 1},           // ON
  {CV_AUX3Servo_speed, 10},    //  10：Slow
  {CV_AUX3Servo_Angle, 100},   // 120 
  {CV_AUX4Servo, 1},           // ONF
  {CV_AUX4Servo_speed, 10},    //  10:Slow
  {CV_AUX4Servo_Angle, 100},   // 120
  {CV_Servo_Delay, 150},       // 150 (150x20= 3000ms) 
  {CV_SWITCH_0, 0},  //Staus of Accesory  First Address  
  {CV_SWITCH_1, 0},  //Staus of Accesory  Second Address
	{200 + 0, 0}, // Function status
	{200 + 1, 0},

};

/*************************************************
 * 
 * Declaration of Functions
 * 
 *************************************************/

void(* resetFunc) (void) = 0;  //declare reset function at address 0
uint8_t FactoryDefaultCVIndex = sizeof(FactoryDefaultCVs) / sizeof(CVPair);
uint16_t gAccessoryAddress = 3;
uint8_t Exec_servo_t(void);
uint8_t Exec_servo_c(void);
/*************************************************
 * 
 * Functions
 * 
 *************************************************/

void notifyCVResetFactoryDefault()
{
	//When anything is writen to CV8 reset to defaults. 

  resetCVToDefault();	
   
  #ifdef DEBUG
    Serial.println("Resetting...");
  #endif
  delay(1000);  //typical CV programming sends the same command multiple times - specially since we dont ACK. so ignore them by delaying
  resetFunc();
}


void resetCVToDefault()
{
	//CVをデフォルトにリセット
  #ifdef DEBUG
	  Serial.println("CVs being reset to factory defaults");
  #endif
	
	for (int j=0; j < FactoryDefaultCVIndex; j++ ){
		Dcc.setCV( FactoryDefaultCVs[j].CV, FactoryDefaultCVs[j].Value);
	}
}

extern void	   notifyCVChange( uint16_t CV, uint8_t Value)
{
   //CVが変更されたときのメッセージ
  #ifdef DEBUG
    Serial.print("CV "); 
    Serial.print(CV); 
    Serial.print(" Changed to "); 
    Serial.println(Value, DEC);
  #endif
}

void notifyCVAck(void)
{
  #ifdef DEBUG
    Serial.println("notifyCVAck");
  #endif
  analogWrite(MOTOR_PWM_B, 255);    //by MP6513.
  analogWrite(MOTOR_PWM_A, 30);     //by MP6513.
  delay( 6 );  
  analogWrite(MOTOR_PWM_A, 255);      //by MP6513.
  analogWrite(MOTOR_PWM_B, 255);      //by MP6513.
}
//**************************************************************************
// This function is called whenever a normal DCC Turnout Packet is received
//**************************************************************************
void notifyDccAccState( uint16_t Addr, uint16_t BoardAddr, uint8_t OutputAddr, uint8_t State)
{
  uint16_t aAccAdress = Addr;
#ifdef DEBUG
    Serial.print("notifyDccAccState: ") ;
    Serial.print(Addr,DEC) ;
    Serial.print(",BAdr:");
    Serial.print(BoardAddr,DEC) ;
    Serial.print(",OAdr:");
    Serial.print(OutputAddr,DEC) ;
    Serial.print(", Target Adr:");
    Serial.print(aAccAdress,DEC) ;
    Serial.print(", gAccessoryAddress:");
    Serial.print(gAccessoryAddress,DEC) ;
    Serial.print(", State:");
    Serial.println(State, HEX) ;
#endif
  //---------------------------------------------------------------------
  //アドレスチェック(CVから得た11bit分の信号を比較)　　//151227 fujigay2
  //----------------------------------------------------------------------
  uint8_t aDir = (OutputAddr & 0b001);
   if ((gAccessoryAddress <= aAccAdress) && ((gAccessoryAddress + 1)  >= aAccAdress ))
   {
      if(aDir  == 0)
      {
        Exec_servo_t( aAccAdress - gAccessoryAddress  + 1 );
      }
      else if( aDir == 1)
      {
        Exec_servo_c( aAccAdress - gAccessoryAddress  + 1 );
      }
   }

}
  
//********************************************************************************
// アドレスの取得
//********************************************************************************
uint16_t getMyAddr_Acc(void)
{
  uint16_t  Addr ;
  uint8_t   CV29Value ;
  CV29Value = Dcc.getCV( CV_29_CONFIG ) ;
    if( CV29Value & 0b10000000 ){  // Accessory Decoder? 
      Addr = ( Dcc.getCV( CV_ACCESSORY_DECODER_ADDRESS_MSB ) << 8 ) | Dcc.getCV( CV_ACCESSORY_DECODER_ADDRESS_LSB ) ;
    }
    else   // Multi-Function Decoder?
    {
      if( CV29Value & 0b00100000 )  // Two Byte Address?
        Addr = ( Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB ) << 8 ) | Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB ) ;

      else
       Addr = Dcc.getCV( 1 ) ;
    }
    
    return Addr ;
}

//**********************************************
//  セットアップ
//**********************************************
void setup()
{
  	//ファンクションの割り当てピン初期化
	  pinMode(PIN_F0_F, OUTPUT);
  	digitalWrite(PIN_F0_F, 0);
  	pinMode(PIN_F0_R, OUTPUT);
    digitalWrite(PIN_F0_R, 1);	
    pinMode(PIN_AUX1, OUTPUT);
    digitalWrite(PIN_AUX1, 0);
	  pinMode(PIN_AUX2, OUTPUT);
	  digitalWrite(PIN_AUX2, 0);
	  pinMode(PIN_AUX5, OUTPUT);
	  digitalWrite(PIN_AUX5, 0);
    pinMode(PIN_AUX6, OUTPUT);
    digitalWrite(PIN_AUX6, 0);
  //PWM出力ピン D3,D11を出力にセット
    pinMode(MOTOR_PWM_B, OUTPUT);
    pinMode(MOTOR_PWM_A, OUTPUT);

  // ソフトウエアシリアル通信レートセット:
    Serial.begin(9600);

  

	  if ( Dcc.getCV(CV_MULTIFUNCTION_PRIMARY_ADDRESS) == 0xFF )
	  {
		   //if eeprom has 0xFF then assume it needs to be programmed
      #ifdef DEBUG
	      Serial.println("CV Defaulting due to blank eeprom");
      #endif
	    notifyCVResetFactoryDefault();
    } 
   else{
      #ifdef DEBUG
	      Serial.println("CV Not Defaulting");
      #endif
    }
	// Setup which External Interrupt, the Pin it's associated with that we're using, disable pullup.
	    Dcc.pin(0, 2, 0);

  // Call the main DCC Init function to enable the DCC Receiver
  Dcc.init( MAN_ID_NUMBER, MAN_VER_NUMBER,  FLAGS_OUTPUT_ADDRESS_MODE | FLAGS_DCC_ACCESSORY_DECODER, 0 );


	//Reset task
	  gPreviousL1 = millis();
	  gPreviousL2 = millis();

  /*  アクセサリアドレス(下位2bitを考慮)の先修得 */
  gAccessoryAddress = getMyAddr_Acc();
  #ifdef DEBUG
  Serial.print("Acc Adr is ");
  Serial.println(gAccessoryAddress);  
  #endif  
	
	//Init CVs
  	gCV1_SAddr = Dcc.getCV( CV_ACCESSORY_DECODER_ADDRESS_LSB ) ;
    gCV9_LAddr = Dcc.getCV( CV_ACCESSORY_DECODER_ADDRESS_MSB ) ;
    gCV61_AUX3Servo =  Dcc.getCV( CV_AUX3Servo ) ;
    gCV62_AUX3Servo_speed =  Dcc.getCV( CV_AUX3Servo_speed ); 
    gCV63_AUX3Servo_Angle =  Dcc.getCV( CV_AUX3Servo_Angle);
    gCV64_AUX4Servo =  Dcc.getCV( CV_AUX4Servo );
    gCV65_AUX4Servo_speed =  Dcc.getCV( CV_AUX4Servo_speed ); 
    gCV66_AUX4Servo_Angle =  Dcc.getCV( CV_AUX4Servo_Angle); 
    gCV69_Servo_Delay = Dcc.getCV( CV_Servo_Delay);
//AUX3,4 pin assignment
    if ( gCV61_AUX3Servo == 1 )
    {
        myservo1.attach(6);
//        myservo1.write( 0 , 0);
    }
    else
    {
        pinMode(PIN_AUX3, OUTPUT);
        digitalWrite(PIN_AUX3, 0);
    }
    if ( gCV64_AUX4Servo == 1 )
   {
       myservo2.attach(5);
 //       myservo2.write( 0 , 0);
    }
    else
    {
       pinMode(PIN_AUX4, OUTPUT);
       digitalWrite(PIN_AUX4, 0);
    }
	/* スイッチ状態テーブル初期化・読み取り */

	  for( int i = 0; i < 2; i++)
	  {
		  gCVFuncTable[i] = Dcc.getCV( CV_SWITCH_0 + i ) ;
		/* 異常値は0に固定 */
  		if( gCVFuncTable[i] > 1)
	  	{
		  	gCVFuncTable[i] = 0;
		  }
      if (( gCV61_AUX3Servo == 1) &&  (gCVFuncTable[i] == 0)){
          myservo1.write( gCV63_AUX3Servo_Angle, gCV62_AUX3Servo_speed); 
          digitalWrite(PIN_AUX1, 0);
          digitalWrite(PIN_AUX2, 1);
      }
      if (( gCV61_AUX3Servo == 1) &&  (gCVFuncTable[i] == 1)){
          myservo1.write( 0 , gCV62_AUX3Servo_speed);
          digitalWrite(PIN_AUX2, 0);
          digitalWrite(PIN_AUX1, 1);
       }
      if (( gCV64_AUX4Servo == 1) &&  (gCVFuncTable[i] == 0)){
         myservo2.write( gCV66_AUX4Servo_Angle, gCV65_AUX4Servo_speed); 
          digitalWrite(PIN_AUX5, 0);
          digitalWrite(PIN_AUX6, 1);
      }
      if (( gCV64_AUX4Servo == 1) &&  (gCVFuncTable[i] == 1)){
          myservo2.write( 0 , gCV65_AUX4Servo_speed); 
          digitalWrite(PIN_AUX6, 0);
          digitalWrite(PIN_AUX5, 1);
      }
	  }
    delay ( gCV69_Servo_Delay * 20 );
    digitalWrite(PIN_F0_R, 0);
}

void loop()
{
	
	// You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
	Dcc.process();
	
}

// Excec Servo motor
void Exec_servo_t(uint8_t Swich_addr)
{
     if (( gCV61_AUX3Servo == 1) && (Swich_addr ==1 )){
         if (gCVFuncTable[0] == 0) {
            myservo1.write( 0 , gCV62_AUX3Servo_speed); 
            digitalWrite(PIN_AUX2, 0);
            digitalWrite(PIN_AUX1, 1);
            gCVFuncTable[0] = 1 ;
            Dcc.setCV(CV_SWITCH_0,gCVFuncTable[0]);
         }
     }  
     if (( gCV64_AUX4Servo == 1 ) && (Swich_addr ==2 )){
         if (gCVFuncTable[1] == 0) {
            myservo2.write( 0 , gCV65_AUX4Servo_speed); 
            digitalWrite(PIN_AUX5, 1);
            digitalWrite(PIN_AUX6, 0);
            gCVFuncTable[1] = 1 ;
            Dcc.setCV(CV_SWITCH_1,gCVFuncTable[1]);
         }
     }  
     //   ServoがOFFの時はAUX3,4に出力
     if (( gCV61_AUX3Servo == 0) && (Swich_addr ==1 )){
         if (gCVFuncTable[0] == 0) {
            digitalWrite(PIN_AUX2, 0);
            digitalWrite(PIN_AUX1, 1);
            digitalWrite(PIN_AUX3, 1);
            gCVFuncTable[0] = 1 ;
            Dcc.setCV(CV_SWITCH_0,gCVFuncTable[0]);
         }
     }
     if (( gCV64_AUX4Servo == 0) && (Swich_addr ==2 )){
         if (gCVFuncTable[1] == 0) {
            digitalWrite(PIN_AUX5, 1);
            digitalWrite(PIN_AUX6, 0);
            digitalWrite(PIN_AUX4, 1);
            gCVFuncTable[1] = 1 ;
            Dcc.setCV(CV_SWITCH_1,gCVFuncTable[1]);
         }
     }

}
void Exec_servo_c(uint8_t Swich_addr)
{
     if (( gCV61_AUX3Servo == 1 ) && (Swich_addr ==1 )){
         if (gCVFuncTable[0] == 1) {
            myservo1.write( gCV63_AUX3Servo_Angle, gCV62_AUX3Servo_speed); 
            digitalWrite(PIN_AUX1, 0);
            digitalWrite(PIN_AUX2, 1);
            gCVFuncTable[0] = 0 ;
            Dcc.setCV(CV_SWITCH_0,gCVFuncTable[0]);
         }   
     }  
     if (( gCV64_AUX4Servo == 1 ) && (Swich_addr ==2 )){
         if (gCVFuncTable[1] == 1) {
            myservo2.write( gCV66_AUX4Servo_Angle, gCV65_AUX4Servo_speed); 
            digitalWrite(PIN_AUX5, 0);
            digitalWrite(PIN_AUX6, 1);
            gCVFuncTable[1] = 0 ;
            Dcc.setCV(CV_SWITCH_1,gCVFuncTable[1]);
         }
     }
     //   ServoがOFFの時はAUX3,4に出力
     if (( gCV61_AUX3Servo == 0) && (Swich_addr ==1 )){
        if (gCVFuncTable[0] == 1) {
           digitalWrite(PIN_AUX1, 0);
           digitalWrite(PIN_AUX2, 1);
           digitalWrite(PIN_AUX3, 0);
            gCVFuncTable[0] = 0 ;
            Dcc.setCV(CV_SWITCH_0,gCVFuncTable[0]);
         } 
     }
     if (( gCV64_AUX4Servo == 0) && (Swich_addr ==2 )){
        if (gCVFuncTable[1] == 1) {
           digitalWrite(PIN_AUX5, 0);
           digitalWrite(PIN_AUX6, 1);
           digitalWrite(PIN_AUX4, 0);
           gCVFuncTable[1] = 0 ;
           Dcc.setCV(CV_SWITCH_1,gCVFuncTable[1]);
        }
     }
}
//		digitalWrite(PIN_F0_F, 1);
//		digitalWrite(PIN_F0_R, 0);	
//		digitalWrite(PIN_F0_F, 0);
// 		digitalWrite(PIN_F0_R, 1);	
 

